#include <windows.h>
#include <stdio.h>
#include <string.h>
#include <io.h>
#include <fcntl.h>
#include <sys/stat.h>

#include "usbspi.h"

#include <setupapi.h>
#include <locale.h>

#pragma comment(lib, "setupapi.lib")

#define MY_DEVICE_ID  "Vid_04d8&Pid_003F"
//  Variables that need to have wide scope.
HANDLE WriteHandle = INVALID_HANDLE_VALUE;	//Need to get a write "handle" to our device before we can write to it.
HANDLE ReadHandle = INVALID_HANDLE_VALUE;	//Need to get a read "handle" to our device before we can read from it.

#ifdef _DEBUG
#define new DEBUG_NEW
#endif

#define PACKET_SIZE		65

unsigned char OutputPacketBuffer[PACKET_SIZE] = {0x00};	//Allocate a memory buffer equal to our endpoint size + 1
unsigned char InputPacketBuffer[PACKET_SIZE] = {0x00};	//Allocate a memory buffer equal to our endpoint size + 1

#define DATA_SIZE_MAX			62

#define SPIADP_CMD_SPI_MODE		0x10	//SPI[h.
#define SPIADP_CMD_SPI_SPEED	0x11	//SPIx.
#define SPIADP_CMD_ACK_SETTING	0x12	//AckON/OFF.
#define SPIADP_CMD_CS_CTRL		0x20	//CS.
#define SPIADP_CMD_GET_INPUT	0x21	//̓s̏Ԏ擾.
#define SPIADP_CMD_SET_OUTPUT	0x22	//o̓s̏Ԑݒ.
#define SPIADP_CMD_GET_OUTPUT	0x23	//o̓s̏Ԏ擾.
#define SPIADP_CMD_DATA			0x30	//f[^ǂݏ.

#define SPIADP_CMND				1
#define SPIADP_LEN				2
#define SPIADP_DATA				3

////////////////////////////////////////////////////////////////////////////
void MakeLower(char *c)
{
	for(; *c != '\0'; *c = tolower(*c), c++);
}

////////////////////////////////////////////////////////////////////////////
void Init()
{
	//-------------------------------------------------------BEGIN CUT AND PASTE BLOCK-----------------------------------------------------------------------------------
	/*
	Before we can "connect" our application to our USB embedded device, we must first find the device.
	A USB bus can have many devices simultaneously connected, so somehow we have to find our device, and only
	our device.  This is done with the Vendor ID (VID) and Product ID (PID).  Each USB product line should have
	a unique combination of VID and PID.

	Microsoft has created a number of functions which are useful for finding plug and play devices.  Documentation
	for each function used can be found in the MSDN library.  We will be using the following functions:

	SetupDiGetClassDevs()					//provided by setupapi.dll, which comes with Windows
	SetupDiEnumDeviceInterfaces()			//provided by setupapi.dll, which comes with Windows
	GetLastError()							//provided by kernel32.dll, which comes with Windows
	SetupDiDestroyDeviceInfoList()			//provided by setupapi.dll, which comes with Windows
	SetupDiGetDeviceInterfaceDetail()		//provided by setupapi.dll, which comes with Windows
	SetupDiGetDeviceRegistryProperty()		//provided by setupapi.dll, which comes with Windows
	malloc()								//part of C runtime library, msvcrt.dll?
	CreateFile()							//provided by kernel32.dll, which comes with Windows

	We will also be using the following unusual data types and structures.  Documentation can also be found in
	the MSDN library:

	PSP_DEVICE_INTERFACE_DATA
	PSP_DEVICE_INTERFACE_DETAIL_DATA
	SP_DEVINFO_DATA
	HDEVINFO
	HANDLE
	GUID

	The ultimate objective of the following code is to call CreateFile(), which opens a communications
	pipe to a specific device (such as a HID class USB device endpoint).  CreateFile() returns a "handle"
	which is needed later when calling ReadFile() or WriteFile().  These functions are used to actually
	send and receive application related data to/from the USB peripheral device.

	However, in order to call CreateFile(), we first need to get the device path for the USB device
	with the correct VID and PID.  Getting the device path is a multi-step round about process, which
	requires calling several of the SetupDixxx() functions provided by setupapi.dll.
	*/


	//Globally Unique Identifier (GUID) for HID class devices.  Windows uses GUIDs to identify things.
	GUID InterfaceClassGuid = { 0x4d1e55b2, 0xf16f, 0x11cf, 0x88, 0xcb, 0x00, 0x11, 0x11, 0x00, 0x00, 0x30 };

	HDEVINFO DeviceInfoTable = INVALID_HANDLE_VALUE;
	PSP_DEVICE_INTERFACE_DATA InterfaceDataStructure = new SP_DEVICE_INTERFACE_DATA;
	PSP_DEVICE_INTERFACE_DETAIL_DATA DetailedInterfaceDataStructure = new SP_DEVICE_INTERFACE_DETAIL_DATA;
	SP_DEVINFO_DATA DevInfoData;

	DWORD InterfaceIndex = 0;
	DWORD StatusLastError = 0;
	DWORD dwRegType;
	DWORD dwRegSize;
	DWORD StructureSize = 0;
	PBYTE PropertyValueBuffer;
	bool MatchFound = false;
	DWORD ErrorStatus;

	//CString DeviceIDToFind = _T(MY_DEVICE_ID);
	char cDeviceIDToFind[] = MY_DEVICE_ID;

	//First populate a list of plugged in devices (by specifying "DIGCF_PRESENT"), which are of the specified class GUID. 
	DeviceInfoTable = SetupDiGetClassDevs(&InterfaceClassGuid, NULL, NULL, DIGCF_PRESENT | DIGCF_DEVICEINTERFACE);

	//Now look through the list we just populated.  We are trying to see if any of them match our device. 
	while (true)
	{
		InterfaceDataStructure->cbSize = sizeof(SP_DEVICE_INTERFACE_DATA);
		if (SetupDiEnumDeviceInterfaces(DeviceInfoTable, NULL, &InterfaceClassGuid, InterfaceIndex, InterfaceDataStructure))
		{
			ErrorStatus = GetLastError();
			if (ERROR_NO_MORE_ITEMS == ErrorStatus)	//Did we reach the end of the list of matching devices in the DeviceInfoTable?
			{	//Cound not find the device.  Must not have been attached.
				SetupDiDestroyDeviceInfoList(DeviceInfoTable);	//Clean up the old structure we no longer need.
				return;
			}
		}
		else	//Else some other kind of unknown error ocurred...
		{
			ErrorStatus = GetLastError();
			SetupDiDestroyDeviceInfoList(DeviceInfoTable);	//Clean up the old structure we no longer need.
			return;
		}

		//Now retrieve the hardware ID from the registry.  The hardware ID contains the VID and PID, which we will then 
		//check to see if it is the correct device or not.

		//Initialize an appropriate SP_DEVINFO_DATA structure.  We need this structure for SetupDiGetDeviceRegistryProperty().
		DevInfoData.cbSize = sizeof(SP_DEVINFO_DATA);
		SetupDiEnumDeviceInfo(DeviceInfoTable, InterfaceIndex, &DevInfoData);

		//First query for the size of the hardware ID, so we can know how big a buffer to allocate for the data.
		SetupDiGetDeviceRegistryProperty(DeviceInfoTable, &DevInfoData, SPDRP_HARDWAREID, &dwRegType, NULL, 0, &dwRegSize);

		//Allocate a buffer for the hardware ID.
		PropertyValueBuffer = (BYTE *)malloc(dwRegSize);
		if (PropertyValueBuffer == NULL)	//if null, error, couldn't allocate enough memory
		{	//Can't really recover from this situation, just exit instead.
			SetupDiDestroyDeviceInfoList(DeviceInfoTable);	//Clean up the old structure we no longer need.
			return;
		}

		//Retrieve the hardware IDs for the current device we are looking at.  PropertyValueBuffer gets filled with a 
		//REG_MULTI_SZ (array of null terminated strings).  To find a device, we only care about the very first string in the
		//buffer, which will be the "device ID".  The device ID is a string which contains the VID and PID, in the example 
		//format "Vid_04d8&Pid_003f".
		SetupDiGetDeviceRegistryProperty(DeviceInfoTable, &DevInfoData, SPDRP_HARDWAREID, &dwRegType, PropertyValueBuffer, dwRegSize, NULL);

		//Now check if the first string in the hardware ID matches the device ID of my USB device.
#ifdef UNICODE
		CString DeviceIDFromRegistry = (wchar_t *)PropertyValueBuffer;
#else
		//CString DeviceIDFromRegistry = (char *)PropertyValueBuffer;
		char cDeviceIDFromRegistry[256];
		strcpy(cDeviceIDFromRegistry, (char *)PropertyValueBuffer);
#endif

		free(PropertyValueBuffer);		//No longer need the PropertyValueBuffer, free the memory to prevent potential memory leaks

										//Convert both strings to lower case.  This makes the code more robust/portable accross OS Versions
										//			DeviceIDFromRegistry = DeviceIDFromRegistry.MakeLower();	
										//			DeviceIDToFind = DeviceIDToFind.MakeLower();
		//DeviceIDFromRegistry.MakeLower();
		//printf("[%s]\n", cDeviceIDFromRegistry);
		MakeLower((char*)&cDeviceIDFromRegistry[0]);
		//printf("[%s]\n", cDeviceIDFromRegistry);

		//DeviceIDToFind.MakeLower();
		//printf("[%s]\n", cDeviceIDToFind);
		MakeLower((char*)&cDeviceIDToFind[0]);
		//printf("[%s]\n", cDeviceIDToFind);

		//Now check if the hardware ID we are looking at contains the correct VID/PID
		//			MatchFound = DeviceIDFromRegistry.Find(DeviceIDToFind);	
		//printf("%s\n", cDeviceIDFromRegistry);
		//printf("%s\n", cDeviceIDToFind);
		//if (-1 != DeviceIDFromRegistry.Find(DeviceIDToFind))
		if (NULL != strstr(cDeviceIDFromRegistry, cDeviceIDToFind))
			//			if(MatchFound != true)
		{
			//Device must have been found.  Open read and write handles.  In order to do this, we will need the actual device path first.
			//We can get the path by calling SetupDiGetDeviceInterfaceDetail(), however, we have to call this function twice:  The first
			//time to get the size of the required structure/buffer to hold the detailed interface data, then a second time to actually 
			//get the structure (after we have allocated enough memory for the structure.)
			DetailedInterfaceDataStructure->cbSize = sizeof(SP_DEVICE_INTERFACE_DETAIL_DATA);
			//First call populates "StructureSize" with the correct value
			SetupDiGetDeviceInterfaceDetail(DeviceInfoTable, InterfaceDataStructure, NULL, NULL, &StructureSize, NULL);
			DetailedInterfaceDataStructure = (PSP_DEVICE_INTERFACE_DETAIL_DATA)(malloc(StructureSize));		//Allocate enough memory
			if (DetailedInterfaceDataStructure == NULL)	//if null, error, couldn't allocate enough memory
			{	//Can't really recover from this situation, just exit instead.
				SetupDiDestroyDeviceInfoList(DeviceInfoTable);	//Clean up the old structure we no longer need.
				return;
			}
			DetailedInterfaceDataStructure->cbSize = sizeof(SP_DEVICE_INTERFACE_DETAIL_DATA);
			//Now call SetupDiGetDeviceInterfaceDetail() a second time to receive the goods.  
			SetupDiGetDeviceInterfaceDetail(DeviceInfoTable, InterfaceDataStructure, DetailedInterfaceDataStructure, StructureSize, NULL, NULL);

			//We now have the proper device path, and we can finally open read and write handles to the device.
			//We store the handles in the global variables "WriteHandle" and "ReadHandle", which we will use later to actually communicate.
			WriteHandle = CreateFile((DetailedInterfaceDataStructure->DevicePath), GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, 0);

			ErrorStatus = GetLastError();
			if (ErrorStatus == ERROR_SUCCESS)
				//					ToggleLED_btn->Enabled = true;				//Make button no longer greyed out
				ReadHandle = CreateFile((DetailedInterfaceDataStructure->DevicePath), GENERIC_READ, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, 0);
			ErrorStatus = GetLastError();
			if (ErrorStatus == ERROR_SUCCESS)
			{
				//					GetPushbuttonState_btn->Enabled = true;		//Make button no longer greyed out
				//					StateLabel->Enabled = true;					//Make label no longer greyed out
			}
			SetupDiDestroyDeviceInfoList(DeviceInfoTable);	//Clean up the old structure we no longer need.
			return;
		}

		InterfaceIndex++;
		//Keep looping until we either find a device with matching VID and PID, or until we run out of items.
	}//end of while(true)	
	 //-------------------------------------------------------END CUT AND PASTE BLOCK-------------------------------------------------------------------------------------
}

////////////////////////////////////////////////////////////////////////////
void spi_send(void)
{
	DWORD BytesWritten = 0;
	DWORD BytesRead = 0;

	WriteFile(WriteHandle, (char*)&OutputPacketBuffer, PACKET_SIZE, &BytesWritten, 0);	//Blocking function, unless an "overlapped" structure is used
	ReadFile(ReadHandle, &InputPacketBuffer, PACKET_SIZE, &BytesRead, 0);		//Blocking function, unless an "overlapped" structure is used	
}

////////////////////////////////////////////////////////////////////////////
void spi_mode(BYTE bMode)
{
	OutputPacketBuffer[SPIADP_CMND] = SPIADP_CMD_SPI_MODE;
	OutputPacketBuffer[SPIADP_LEN]  = 1;
	OutputPacketBuffer[SPIADP_DATA]  = bMode;

	spi_send();
}

////////////////////////////////////////////////////////////////////////////
void spi_speed(BYTE bSpeed)
{
	OutputPacketBuffer[SPIADP_CMND] = SPIADP_CMD_SPI_SPEED;
	OutputPacketBuffer[SPIADP_LEN]  = 1;
	OutputPacketBuffer[SPIADP_DATA]  = bSpeed;

	spi_send();
}

////////////////////////////////////////////////////////////////////////////
void spi_ack(BYTE bAckFlg)
{
	OutputPacketBuffer[SPIADP_CMND] = SPIADP_CMD_ACK_SETTING;
	OutputPacketBuffer[SPIADP_LEN]  = 1;
	OutputPacketBuffer[SPIADP_DATA]  = bAckFlg;

	spi_send();
}

////////////////////////////////////////////////////////////////////////////
void spi_cs(BYTE bFlg)
{
	OutputPacketBuffer[SPIADP_CMND] = SPIADP_CMD_CS_CTRL;
	OutputPacketBuffer[SPIADP_LEN]  = 1;
	OutputPacketBuffer[SPIADP_DATA]  = bFlg;

	spi_send();
}

////////////////////////////////////////////////////////////////////////////
BYTE spi_get_input_pin(void)
{
	OutputPacketBuffer[SPIADP_CMND] = SPIADP_CMD_GET_INPUT;
	OutputPacketBuffer[SPIADP_LEN]  = 0;

	spi_send();

	return InputPacketBuffer[SPIADP_DATA];
}

////////////////////////////////////////////////////////////////////////////
void spi_set_output_pin(BYTE bFlg)
{
	OutputPacketBuffer[SPIADP_CMND] = SPIADP_CMD_SET_OUTPUT;
	OutputPacketBuffer[SPIADP_LEN]  = 1;
	OutputPacketBuffer[SPIADP_DATA]  = bFlg;

	spi_send();
}

////////////////////////////////////////////////////////////////////////////
BYTE spi_get_output_pin(void)
{
	OutputPacketBuffer[SPIADP_CMND] = SPIADP_CMD_GET_OUTPUT;
	OutputPacketBuffer[SPIADP_LEN]  = 0;

	spi_send();

	return InputPacketBuffer[SPIADP_DATA];
}

////////////////////////////////////////////////////////////////////////////
BYTE spi_rw(BYTE bLen, BYTE *pbSndDat, BYTE *pbRcvDat)
{
	if(bLen > DATA_SIZE_MAX){
		return 0xff;
	}

	OutputPacketBuffer[SPIADP_CMND] = SPIADP_CMD_DATA;
	OutputPacketBuffer[SPIADP_LEN]  = bLen;
	memcpy((void*)&OutputPacketBuffer[SPIADP_DATA], pbSndDat, bLen);

	spi_send();

	memcpy(pbRcvDat, (void*)&InputPacketBuffer[SPIADP_DATA], bLen);

	return bLen;
}

////////////////////////////////////////////////////////////////////////////
BYTE spi_rw_single(BYTE bDat)
{
	OutputPacketBuffer[SPIADP_CMND] = SPIADP_CMD_DATA;
	OutputPacketBuffer[SPIADP_LEN]  = 1;
	OutputPacketBuffer[SPIADP_DATA]  = bDat;

	spi_send();

	return InputPacketBuffer[SPIADP_DATA];
}

